/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#ifndef __CARD_H
#define __CARD_H

#include "compil.h"

typedef struct {
	/* Horizontal CRTC */
	int HTotal;
	int HDisp;
	int HBStart;
	int HSStart;
	int HSEnd;
	int HBEnd;

	/* Vertical CRTC, doubled in doublescan, halved in interlaced */
	int VTotal;
	int VDisp;
	int VBStart;
	int VSStart;
	int VSEnd;
	int VBEnd;

	int width; /* Size of the mode */
	int height;
	int original_width; /* Original size of the mode */
	int original_height;

	int hpolarity;
	int vpolarity;
	int doublescan; /* Doublescan flag */
	int interlace; /* Interlace flag */
	int interlaceratio; /* Interlace ratio, 0-100 */
	int bits; /* Bits per pixel */

	/* Clocks */
	long dotclockHz;
} CRTCPARAM;

void SetCrtcVGARegister(CRTCPARAM STACK_PTR *cp);
void SetGenericVGARegister(CRTCPARAM STACK_PTR *cp);

DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos);
void _out(int edx, BYTE al);
BYTE _in(int edx);
void _outb(int edx, BYTE al);
BYTE _inb(int edx);
void _outw(int edx, WORD ax);
WORD _inw(int edx);
void _outl(int edx, DWORD eax);
DWORD _inl(int edx);

int GetCrtcAddress(void);
BYTE GetSeqReg(BYTE);
void SetSeqReg(BYTE, BYTE);
BYTE GetCrtcReg(BYTE);
void SetCrtcReg(BYTE, BYTE);
BYTE GetAtrbData(BYTE);
void SetAtrbData(BYTE, BYTE);
BYTE GetGraphReg(BYTE);
void SetGraphReg(BYTE, BYTE);
int GetSeqDivider(void);
void SetSeqDivider(BYTE);
void SetHVPolarity(int HPolar, int VPolar);
void SetDoubleScan(int flag);
void DisableGenerateSignal(void);
void EnableGenerateSignal(void);
int GetMode15KHz(CRTCPARAM STACK_PTR *cp);
void VGAUnlock(void);
void VGALock(void);

/***************************************************************************/
/* Log */

#ifdef __REAL__

#define CARD_LOG(a) do { } while (0)

#else

void card_log(const char *text,...);

#define CARD_LOG(a) card_log a

#endif

/***************************************************************************/
/* PCI */

int pci_detect(void);
int pci_find_device(unsigned vendor, unsigned device, unsigned STACK_PTR* bus_device_func);
int pci_read_dword(unsigned bus_device_func, unsigned reg, DWORD STACK_PTR* value);
int pci_write_dword(unsigned bus_device_func, unsigned reg, DWORD value);

int pci_BIOS_address_map(unsigned bus_device_func);
void pci_BIOS_address_unmap(void);
void pci_BIOS_read(void STACK_PTR* dest, unsigned offset, unsigned len);

int pci_MMIO_address_map(unsigned bus_device_func, unsigned reg, DWORD mask);
void pci_MMIO_address_unmap(void);
int pci_MMIO_address_relocate(void);
void pci_MMIO_address_restore(void);
unsigned pci_MMIO_selector_get(void);

/***************************************************************************/
/* Mapping */

#define _create_linear_mapping card_create_linear_mapping
#define _remove_linear_mapping card_remove_linear_mapping
#define _create_selector card_create_selector
#define _remove_selector card_remove_selector

extern int card_create_linear_mapping(unsigned long *linear, unsigned long physaddr, int size);
extern int card_create_selector(unsigned *segment, unsigned long linear, int size);
extern void card_remove_linear_mapping(unsigned long *linear);
extern void card_remove_selector(unsigned *segment);

#endif
