/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include "compil.h"
#include "matrox.h"
#include "card.h"

#define USE_ZOOM
//#define ZOOMLIMITATION

static void SetMatroxExtRegister(CRTCPARAM *cp);
static void SetMatroxInterlace(int flag, int HSyncStart, int HSyncEnd, int HTotal, int ratio);
static long GetMatroxVclkParam(long dotclock, int STACK_PTR* s_mul, int STACK_PTR* s_div, int STACK_PTR* s_p, int STACK_PTR* s_s);
static void SetMatroxVClock(int N, int M, int P, int S);
static void SetMatroxHZoom(int d0);
static int GetMatroxHorizontalZoomRatio(CRTCPARAM STACK_PTR* cp);
static void AdjustMatroxCrtcParam(CRTCPARAM STACK_PTR* cp, int hzoom);
#ifdef ZOOMLIMITATION
static int CheckMatroxHorizontalTotal(int HTotal, int hzoom);
#endif
static void SetMatroxDoubleScan(int flag);
static DWORD GetMatroxBytePerLine(void);
static void SetMatroxBytePerLine(DWORD d0);

static BYTE MatroxIdxReg_inb(BYTE _index);
static BYTE MatroxCrtcExtReg_inb(BYTE _index);
static void MatroxIdxReg_outb(BYTE _index, BYTE _data);
static void MatroxCrtcExtReg_outb(BYTE _index, BYTE _data);

struct MatroxID {
	unsigned value;
	const char* name;
	int supported;
	int mul_min;
	int mul_max;
	int div_min;
	int div_max;
	int p_min;
	int p_max;
	long ref; /* kHz */
	long vco_min; /* kHz */
	long vco_max; /* kHz */
};

/* Configuration */
#define MATROX_CONF_UNSUPPORTED 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define MATROX_CONF_MYSTIQUE_ID1 1, 101, 128, 2, 32, 0, 3, 14318, 50000, 135000
#define MATROX_CONF_MYSTIQUE_ID2 1, 101, 128, 2, 32, 0, 3, 14318, 50000, 175000
#define MATROX_CONF_MYSTIQUE_ID3 1, 101, 128, 2, 32, 0, 3, 14318, 50000, 220000
#define MATROX_CONF_G100 1, 8, 128, 2, 7, 0, 3, 27050, 50000, 135000
#define MATROX_CONF_G200 1, 8, 128, 2, 7, 0, 3, 27050, 50000, 230000
#define MATROX_CONF_G400 1, 8, 128, 2, 32, 0, 3, 27050, 50000, 300000

static struct MatroxID MatroxIDList[] = {
{ 0x0518, "Matrox MGA 2085PX", MATROX_CONF_UNSUPPORTED },
{ 0x0519, "Matrox Millenium 2064W", MATROX_CONF_UNSUPPORTED },
{ 0x051A, "Matrox Mystique 1064SG", MATROX_CONF_MYSTIQUE_ID1 },
{ 0x051B, "Matrox Millenium 2 2164W", MATROX_CONF_UNSUPPORTED },
{ 0x051F, "Matrox Millenium 2 AGP 2164W", MATROX_CONF_UNSUPPORTED },
{ 0x0D10, "Matrox Impression", MATROX_CONF_UNSUPPORTED },
{ 0x0520, "Matrox G200 PCI", MATROX_CONF_G200 },
{ 0x0521, "Matrox G200 AGP", MATROX_CONF_G200 }, /* TESTED */
{ 0x0525, "Matrox G400 AGP", MATROX_CONF_G400 }, /* TESTED */
{ 0x1000, "Matrox G100 PCI", MATROX_CONF_G100 },
{ 0x1001, "Matrox G100 AGP", MATROX_CONF_G100 },
{ 0, 0, MATROX_CONF_UNSUPPORTED }
};

static unsigned MatroxBusDeviceFunc;

static struct MatroxID* matrox_card; /* Matrox detected */

const char* driverMatrox(void) {
	return matrox_card->name;
}

int detectMatrox(void)
{
	int i;

#if 0 /* fake detect */
	matrox_card = MatroxIDList + 8;
	return 1;
#endif

	if (pci_detect()!=0) {
		CARD_LOG(("matrox: PCI BIOS not Installed.\n"));
		return 0;
	}

	for(i=0;MatroxIDList[i].name;++i) {
		if (pci_find_device(0x0000102B,MatroxIDList[i].value,&MatroxBusDeviceFunc)==0)
			break;
	}

	if(!MatroxIDList[i].name){
		CARD_LOG(("matrox: not found\n"));
		return 0;
	}

	CARD_LOG(("matrox: found %s, device id %04x\n", MatroxIDList[i].name, MatroxIDList[i].value));

	if (!MatroxIDList[i].supported) {
		CARD_LOG(("matrox: card not supported\n"));
		return 0;
	}

	matrox_card = MatroxIDList + i;

	if (pci_MMIO_address_map(MatroxBusDeviceFunc,0x14,0xFFFFC000)!=0) {
		CARD_LOG(( "matrox: pci_address_map error\n"));
		return 0;
	}

	return 1;
}

void resetMatrox(void)
{
	pci_MMIO_address_unmap();
}

int setMatrox(void)
{
	STACK_DECL CRTCPARAM cp;
	int N, M, P, S;
	int hzoom;

	if(GetMode15KHz(&cp) == 0) return 0;

	hzoom = GetMatroxHorizontalZoomRatio(&cp);
	CARD_LOG(("matrox: horizontal zoom ratio %d\n",hzoom));
	AdjustMatroxCrtcParam(&cp, hzoom);

	CARD_LOG(("matrox: clock requested:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));
	cp.dotclockHz = GetMatroxVclkParam(cp.dotclockHz, &N, &M, &P, &S);
	P = 1 << P; /* GetMatroxVclkParam compute P as the number of shift, not as a divisor */
	if (cp.dotclockHz<0) {
		CARD_LOG(("matrox: clock out of range\n"));
		return 0;
	}
	CARD_LOG(("matrox: clock selected:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));

	if (pci_MMIO_address_relocate()!=0) {
		CARD_LOG(( "matrox: pci_address_relocate error\n"));
		return 0;
	}

	DisableGenerateSignal();
	SetGenericVGARegister(&cp);
	SetMatroxExtRegister(&cp);
	SetMatroxVClock(N, M, P, S);
	SetMatroxHZoom(hzoom);
	SetMatroxDoubleScan(cp.doublescan);
	SetHVPolarity(cp.hpolarity, cp.vpolarity);
	EnableGenerateSignal();

	pci_MMIO_address_restore();

	return 1;
}

static int GetMatroxHorizontalZoomRatio(CRTCPARAM STACK_PTR* cp)
{
	int zoom = 1;
#ifdef USE_ZOOM
	if(cp->HTotal < 512){
		zoom = 4;
	}else if(cp->HTotal < 1024){
		zoom = 2;
	}else{
		zoom = 1;
	}

#endif
	return zoom;
}

static void AdjustMatroxCrtcParam(CRTCPARAM STACK_PTR* cp, int hzoom)
{
	cp->HTotal *= hzoom;
	cp->HDisp *= hzoom;
	cp->HBStart *= hzoom;
	cp->HBEnd *= hzoom;
	cp->HSStart *= hzoom;
	cp->HSEnd *= hzoom;
	cp->dotclockHz *= hzoom;

	cp->HBStart = cp->HDisp - 8;
	cp->HBEnd = cp->HTotal - 8;
	cp->VBStart = cp->VDisp  - 1;
	cp->VBEnd = cp->VTotal - 1;

	if( (cp->HBEnd - cp->HBStart) > (0x7F << 3) ){
		cp->HBEnd = cp->HBStart + (0x7F << 3);
	}
	if( (cp->HSEnd - cp->HSStart) > (0x1F << 3) ){
		cp->HSEnd = cp->HSStart + (0x1F << 3);
	}

	if(cp->interlace != 0){
		if( (cp->VTotal % 2) != 0 ){
			cp->VTotal += 1;
			cp->VBEnd       = cp->VTotal - 1;
		}
	}

#ifdef ZOOMLIMITATION
	if( CheckMatroxHorizontalTotal(cp->HTotal, hzoom) == 0 ){
		cp->HTotal += 8;
	}
#endif
}

#ifdef ZOOMLIMITATION
static int CheckMatroxHorizontalTotal(int HTotal, int hzoom)
{
	int result = 1;
	int d0;

	d0 = 8 * hzoom;
	if( ((HTotal >> 3) % d0)  ==  (d0 - 1) ){
		result = 0;
	}
	return result;
}
#endif

static void SetMatroxInterlace(int flag, int HSyncStart, int HSyncEnd, int HTotal, int ratio)
{
	int d1;

	if(flag != 0){
		d1 = MatroxCrtcExtReg_inb(0x00);
		if((d1 & 0x80) == 0){
			d1 = (HSyncStart/8) + (HSyncEnd/8) - (HTotal/8);
			d1 = (d1 * ratio) / 100;
			d1 = (d1 / 2) - 1;
			if(d1 < 0)        d1 = 0;
			if(d1 > 0xFF) d1 = 0xFF;
			MatroxCrtcExtReg_outb(0x05, d1 & 0xFF);
			MatroxCrtcExtReg_outb(0x00, BitMov(MatroxCrtcExtReg_inb(0x00), 7, 1, 0));
			SetMatroxBytePerLine(GetMatroxBytePerLine() * 2);
		}
	}else{
		d1 = MatroxCrtcExtReg_inb(0x00);
		if((d1 & 0x80) != 0){
			MatroxCrtcExtReg_outb(0x00, BitMov(MatroxCrtcExtReg_inb(0x00), 7, 0, 0));
			MatroxCrtcExtReg_outb(0x05, 0);
			SetMatroxBytePerLine(GetMatroxBytePerLine() / 2);
		}
	}
}

static DWORD GetMatroxBytePerLine(void)
{
	DWORD d0;

	d0 = GetCrtcReg(0x13);
	d0 = BitMov(d0, 8, MatroxCrtcExtReg_inb(0x00), 4);
	d0 = BitMov(d0, 9, MatroxCrtcExtReg_inb(0x00), 5);
	return d0;
}

static void SetMatroxBytePerLine(DWORD d0)
{
	SetCrtcReg(0x13, d0 & 0xFF);
	MatroxCrtcExtReg_outb(0x00, BitMov(MatroxCrtcExtReg_inb(0x00), 4, d0, 8));
	MatroxCrtcExtReg_outb(0x00, BitMov(MatroxCrtcExtReg_inb(0x00), 5, d0, 9));
}

static void SetMatroxExtRegister(CRTCPARAM *cp)
{
	int d0;

	d0 = 0;
	d0 = BitMov(d0, 0, (cp->HTotal  >> 3) - 5, 8);
	d0 = BitMov(d0, 1, (cp->HBStart >> 3)    , 8);
	d0 = BitMov(d0, 2, (cp->HSStart >> 3)    , 8);
	d0 = BitMov(d0, 6, (cp->HBEnd   >> 3)    , 6);
	MatroxCrtcExtReg_outb(0x01, d0);

	d0 = 0;
	d0 = BitMov(d0, 0, cp->VTotal - 2, 10);
	d0 = BitMov(d0, 1, cp->VTotal - 2, 11);
	d0 = BitMov(d0, 2, cp->VDisp  - 1, 10);
	d0 = BitMov(d0, 3, cp->VBStart   , 10);
	d0 = BitMov(d0, 4, cp->VBStart   , 11);
	d0 = BitMov(d0, 5, cp->VSStart   , 10);
	d0 = BitMov(d0, 6, cp->VSStart   , 11);
	MatroxCrtcExtReg_outb(0x02, d0);

	SetMatroxInterlace(cp->interlace, cp->HSStart, cp->HSEnd, cp->HTotal, cp->interlaceratio);
}

static long GetMatroxVclkParam(long dotclock, int STACK_PTR* s_mul, int STACK_PTR* s_div, int STACK_PTR* s_p, int STACK_PTR* s_s)
{
	int best_mul,best_div,best_p;
	long best_clock,best_vco;
	int pll_mul,pll_div,pll_p;

	CARD_LOG(("matrox: clock range is %.2f - %.2f MHz\n", (double)matrox_card->vco_min / 1000 / (1 << matrox_card->p_max), (double)matrox_card->vco_max / 1000 / (1 << matrox_card->p_min) ));

	best_clock = -1;
	best_mul = 0;
	best_div = 0;
	best_p = 0;
	best_vco = 0;

	for(pll_div=matrox_card->div_min;pll_div<=matrox_card->div_max;++pll_div)
	for(pll_p=matrox_card->p_min;pll_p<=matrox_card->p_max;++pll_p)
	for(pll_mul=matrox_card->mul_min;pll_mul<=matrox_card->mul_max;++pll_mul) {
		long vco = matrox_card->ref * 1000L * pll_mul / pll_div;
		long new_clock = vco / (1 << pll_p);

		if (vco < matrox_card->vco_min*1000L && pll_p!=matrox_card->p_max)
			continue;
		if (vco > matrox_card->vco_max*1000L)
			continue;

		if (best_clock < 0 || labs(new_clock - dotclock) < labs(best_clock - dotclock)) {
			best_mul = pll_mul;
			best_div = pll_div;
			best_p = pll_p;
			best_clock = new_clock;
			best_vco = vco;
		}
	}

	if (best_clock < 0)
		return -1;

	if (best_vco < matrox_card->vco_min*1000)
		CARD_LOG(("matrox: VCO clock %.2f is OUT of range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)matrox_card->vco_min / 1000, (double)matrox_card->vco_max / 1000));
	else
		CARD_LOG(("matrox: VCO clock %.2f is in range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)matrox_card->vco_min / 1000, (double)matrox_card->vco_max / 1000));

	*s_mul = best_mul;
	*s_div = best_div;
	*s_p = best_p;

	CARD_LOG(("matrox: ref:%.2f, mul:%d, div:%d, shift:%d\n",(double)matrox_card->ref / 1000,best_mul,best_div,best_p));

	/* From XFree 4.0.0 */
	if (best_vco/1000 < 100000)
		*s_s = 0;
	else if (best_vco/1000 < 140000)
		*s_s = 1;
	else if (best_vco/1000 < 180000)
		*s_s = 2;
	else
		*s_s = 3;

	return best_clock;
}

static void SetMatroxVClock(int N, int M, int P, int S)
{
	int PLLn;
	BYTE d0;

	MatroxIdxReg_outb(0x1A, BitMov(MatroxIdxReg_inb(0x1A), 2, 1, 0));

	PLLn = _in(0x3CC);
	PLLn = PLLn | 0x0C;
	_out(0x3C2, PLLn);

	M -= 1;
	N -= 1;
	P -= 1;

	d0 = M & 0x1F;
	MatroxIdxReg_outb(0x4C, d0);
	d0 = N & 0x7F;
	MatroxIdxReg_outb(0x4D, d0);
	d0 = (P & 0x7) | ((S & 0x3) << 3);
	MatroxIdxReg_outb(0x4E, d0);

#if 0
	// TODO Remove ? XFree doesn't do anything
	while(1){
		if( (MatroxIdxReg_inb(0x4F) & 0x40) != 0 ) break;
	}
#endif

	MatroxIdxReg_outb(0x1A, BitMov(MatroxIdxReg_inb(0x1A), 2, 0, 0));
}

static void SetMatroxHZoom(int d0)
{
	if(d0 == 1){
		d0 = 0;
	}else if(d0 == 2){
		d0 = 1;
	}else if(d0 == 4){
		d0 = 3;
	}else{
		d0 = 0;
	}
	MatroxIdxReg_outb(0x38, d0);
}

static void SetMatroxDoubleScan(int flag)
{
	int d0;

	d0 = GetCrtcReg(0x09) & ~0x1F;
	if(flag != 0){
		d0 = d0 | 0x01;
	}
	SetCrtcReg(0x09, d0);
}

static BYTE MatroxIdxReg_inb(BYTE _index)
{
	_farpokeb(pci_MMIO_selector_get(), 0x3C00, _index);
	return _farpeekb(pci_MMIO_selector_get(), 0x3C0A);
}

static void MatroxIdxReg_outb(BYTE _index, BYTE _data)
{
	_farpokeb(pci_MMIO_selector_get(), 0x3C00, _index);
	_farpokeb(pci_MMIO_selector_get(), 0x3C0A, _data);
}

static BYTE MatroxCrtcExtReg_inb(BYTE _index)
{
	_farpokeb(pci_MMIO_selector_get(), 0x1FDE, _index);
	return _farpeekb(pci_MMIO_selector_get(), 0x1FDF);
}

static void MatroxCrtcExtReg_outb(BYTE _index, BYTE _data)
{
	_farpokeb(pci_MMIO_selector_get(), 0x1FDE, _index);
	_farpokeb(pci_MMIO_selector_get(), 0x1FDF, _data);
}
