/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include "compil.h"
#include "s3.h"
#include "card.h"

static void SetS3ExtRegister(CRTCPARAM STACK_PTR *cp);
static void SetS3Interlace(int flag, CRTCPARAM STACK_PTR* cp);
static void SetS3VClock(int pll_mul, int pll_div, int pll_p);
static long GetS3VclkParam(long dotclock, int STACK_PTR *s_mul, int STACK_PTR *s_div, int STACK_PTR *s_p);static void SetS3DoubleScan(int flag);
static void SetS3PLL(void);

struct S3ID {
	unsigned value;
	const char* name;
	int supported;
	int mul_min;
	int mul_max;
	int div_min;
	int div_max;
	int p_min;
	int p_max;
	long ref; /* kHz */
	long vco_min; /* kHz */
	long vco_max; /* kHz */
	int clock_type;
};

/* Clock */
#define S3_CLOCK_TRIO 1
#define S3_CLOCK_VIRGEGX2 2
#define S3_CLOCK_AURORA 3

/* Configuration */
#define S3_CONF_UNSUPPORTED 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define S3_CONF_TRIO 1, 1+2, 127+2, 1+2, 31+2, 0, 3, 14318, 135000, 270000, S3_CLOCK_TRIO
#define S3_CONF_TRIO64V2 1, 1+2, 127+2, 1+2, 31+2, 0, 4, 14318, 170000, 340000, S3_CLOCK_TRIO
#define S3_CONF_VIRGEDXGX 1, 1+2, 127+2, 1+2, 31+2, 0, 4, 14318, 135000, 270000, S3_CLOCK_TRIO
#define S3_CONF_VIRGEVX 1, 1+2, 127+2, 1+2, 31+2, 0, 4, 14318, 220000, 440000, S3_CLOCK_TRIO
#define S3_CONF_VIRGEGX2 1, 1+2, 127+2, 1+2, 31+2, 0, 4, 14318, 170000, 340000, S3_CLOCK_VIRGEGX2
#define S3_CONF_TRIO3D 1, 1+2, 127+2, 1+2, 31+2, 0, 4, 14318, 230000, 460000, S3_CLOCK_TRIO
#define S3_CONF_AURORA 1, 1+2, 127+2, 1+2, 63+2, 0, 3, 14318, 135000, 270000, S3_CLOCK_AURORA

static struct S3ID S3IDList[] = {
{ 0x0551, "S3 PLATO/PX", S3_CONF_UNSUPPORTED },
{ 0x5631, "S3 86C325 ViRGE 3D GUI Accelerator", S3_CONF_TRIO },
{ 0x8800, "S3 Vision 866", S3_CONF_TRIO },
{ 0x8801, "S3 Vision 964", S3_CONF_TRIO },
{ 0x8810, "S3 Trio32", S3_CONF_TRIO },
{ 0x8811, "S3 Trio64, or Trio64V+", S3_CONF_TRIO },
{ 0x8812, "S3 Aurora64V+", S3_CONF_AURORA },
{ 0x8814, "S3 Trio64UV+", S3_CONF_TRIO },
{ 0x883D, "S3 ViRGE/VX", S3_CONF_VIRGEVX },
{ 0x8880, "S3 868", S3_CONF_TRIO },
{ 0x88B0, "S3 928", S3_CONF_TRIO },
{ 0x88C0, "S3 864-1", S3_CONF_TRIO },
{ 0x88C1, "S3 864-2", S3_CONF_TRIO },
{ 0x88C2, "S3 864-3", S3_CONF_TRIO },
{ 0x88C3, "S3 864-4", S3_CONF_TRIO },
{ 0x88D0, "S3 964-1", S3_CONF_TRIO },
{ 0x88D1, "S3 964-2", S3_CONF_TRIO },
{ 0x88D2, "S3 964-3", S3_CONF_TRIO },
{ 0x88D3, "S3 964-4", S3_CONF_TRIO },
{ 0x88F0, "S3 968", S3_CONF_TRIO },
{ 0x88F1, "S3 968-2", S3_CONF_TRIO },
{ 0x88F2, "S3 968-3", S3_CONF_TRIO },
{ 0x88F3, "S3 968-3", S3_CONF_TRIO },
{ 0x8900, "S3 Trio64 V2/DX GUI Accelerator", S3_CONF_TRIO },
{ 0x8901, "S3 Trio64V2", S3_CONF_TRIO64V2 }, /* TESTED */
{ 0x8902, "S3 PLATO/PX", S3_CONF_UNSUPPORTED },
{ 0x8904, "S3 Trio3D", S3_CONF_TRIO3D },
{ 0x8A13, "S3 Trio3D/2X", S3_CONF_TRIO3D },
{ 0x8A01, "S3 ViRGE/DX or ViRGE/GX", S3_CONF_VIRGEDXGX }, /* TESTED */
{ 0x8A10, "S3 ViRGE/GX2", S3_CONF_VIRGEGX2  },
{ 0x8A20, "S3 Savage3D", S3_CONF_UNSUPPORTED },
{ 0x8A21, "S3 Savage3D MacroVision", S3_CONF_UNSUPPORTED },
{ 0x8A22, "S3 Savage4", S3_CONF_UNSUPPORTED },
{ 0x9102, "S3 Savage 2000", S3_CONF_UNSUPPORTED },
{ 0x8C01, "S3 ViRGE/MX", S3_CONF_UNSUPPORTED  },
{ 0x8C02, "S3 ViRGE/MX+", S3_CONF_UNSUPPORTED  },
{ 0x8C03, "S3 ViRGE/MX+MV", S3_CONF_UNSUPPORTED  },
{ 0xca00, "S3 SonicVibes,", S3_CONF_UNSUPPORTED },
{ 0,0, S3_CONF_UNSUPPORTED }
};

static unsigned S3BusDeviceFunc;

static struct S3ID* s3_card; /* S3 detected */

const char* driverS3(void) {
	return s3_card->name;
}

int detectS3(void)
{
	int i;

	if (pci_detect()!=0) {
		CARD_LOG(("s3: PCI BIOS not Installed.\n"));
		return 0;
	}

	for(i=0;S3IDList[i].name;++i) {
		if (pci_find_device(0x00005333,S3IDList[i].value,&S3BusDeviceFunc)==0)
			break;
	}

	if(!S3IDList[i].name){
		CARD_LOG(("s3: not found\n"));
		return 0;
	}

	CARD_LOG(("s3: found %s, device id %04x\n", S3IDList[i].name, S3IDList[i].value));

	if (!S3IDList[i].supported) {
		CARD_LOG(("s3: card not supported\n"));
		return 0;
	}

	s3_card = S3IDList + i;

	return 1;
}

void resetS3(void)
{
}

int setS3(void)
{
	STACK_DECL CRTCPARAM cp;
	int pll_mul,pll_div,pll_p;
	unsigned d0;

	if(GetMode15KHz(&cp) == 0) return 0;

	d0 = GetCrtcReg(0x67);
	if ( (d0 & 0xF0) == (5 << 4) || (d0 & 0xF0) == (3 << 4)) {
		CARD_LOG(("s3: double h value for 16 bit mode\n"));
		cp.HDisp *= 2;
		cp.HTotal *= 2;
		cp.HBStart *= 2;
		cp.HBEnd *= 2;
		cp.HSStart *= 2;
		cp.HSEnd *= 2;
	}

	CARD_LOG(("s3: clock requested:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));
	cp.dotclockHz = GetS3VclkParam(cp.dotclockHz, &pll_mul, &pll_div, &pll_p);
	if (cp.dotclockHz < 0) {
		CARD_LOG(("s3: clock out of range\n"));
		return 0;
	}
	CARD_LOG(("s3: clock selected:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));

	DisableGenerateSignal();

	/* Unlock all */
	SetSeqReg(0x08, 0x06); /* PLL Unlock */
	/* VERT timing regs (CRTC index 6,7(bit0,2,3,5,7),9,10,11(bits0..3),15,16 ) */
	/* HOR timing regs (CRTC index 0..5, 17(bit2) ) */
	/* bit 0 of Clocking mode reg unlocked (8/9 dot font selection) */
	/* Clock bits in MISC reg unlocked */
	SetCrtcReg(0x35, GetCrtcReg(0x35) & 0x0F);
	SetCrtcReg(0x38, 0x48); /* S3 register set (CRTC index 0x30..0x3C) */
	SetCrtcReg(0x39, 0xA5); /* system extension regs (CRTC index 0x50..0x5E) */

	SetCrtcVGARegister(&cp);
	SetS3ExtRegister(&cp);
	SetS3VClock(pll_mul, pll_div, pll_p);
	EnableGenerateSignal();

	return 1;
}

static void SetS3PLL(void)
{
	BYTE d0;

	d0 = _in(0x3CC);
	d0 = d0 | 0x0C;
	_out(0x3C2, d0);
}

static void SetS3ExtRegister(CRTCPARAM STACK_PTR *cp)
{
	int d0;
	int d1;
	int HTotal;
	int HSStart;

	HTotal  = cp->HTotal  >> 3;
	HSStart = cp->HSStart >> 3;

	d1 = ((HTotal - 5) + HSStart) / 2;
	if((d1 - HSStart) < 4){
	if ((HSStart + 4) <= (HTotal - 5))
		d1 = HSStart + 4;
	else
		d1 = HTotal - 5 + 1;
	}

	d0 = 0;
	d0 = BitMov(d0, 0, (cp->HTotal  >> 3) - 5, 8);
	d0 = BitMov(d0, 1, (cp->HDisp   >> 3) - 1, 8);
	d0 = BitMov(d0, 2, (cp->HBStart >> 3)    , 8);
	d0 = BitMov(d0, 4, (cp->HSStart >> 3)    , 8);
	if ((cp->HBEnd >> 3) - (cp->HBStart >> 3) > 64)
		d0 = BitMov(d0, 3, 1, 0);
	if ((cp->HSEnd >> 3) - (cp->HSStart >> 3) > 32)
		d0 = BitMov(d0, 5, 1, 0);
	d0 = BitMov(d0, 6, d1, 8);

	SetCrtcReg(0x5D, d0);                                           // ext horz reg
	SetCrtcReg(0x3B, d1);                                           // data transfer exec pos
	CARD_LOG(("s3: data transfer exec pos %d\n", d1 << 3));

	d0 = 0;
	d0 = BitMov(d0, 0, cp->VTotal  - 2, 10);
	d0 = BitMov(d0, 1, cp->VDisp   - 1, 10);
	d0 = BitMov(d0, 2, cp->VBStart,         10);
	d0 = BitMov(d0, 4, cp->VSStart,         10);
	d0 = BitMov(d0, 6, 1, 0);                                       // line compare
	SetCrtcReg(0x5E, d0);                                           // ext vert reg

	SetS3DoubleScan(cp->doublescan);
	SetS3Interlace(cp->interlace, cp);
	SetHVPolarity(cp->hpolarity, cp->vpolarity);
}

static void SetS3DoubleScan(int flag)
{
	BYTE d0 = GetCrtcReg(0x09);
	if(flag == 0){
		SetCrtcReg(0x09, d0 & 0xE0);
	}else{
		SetCrtcReg(0x09, (d0 & 0xE0) | 0x80);
	}
}

static void SetS3Interlace(int flag, CRTCPARAM STACK_PTR *cp)
{
	int d1;

	if(flag == 0){
		d1 = GetCrtcReg(0x42) & ~0x20;
		SetCrtcReg(0x42, d1);
	}else{
		d1 = ((cp->HTotal >> 3) - 5) / 2;
		d1 = (d1 * cp->interlaceratio) / 100;
		SetCrtcReg(0x3C, d1);                                           // interlace offset
		d1 = GetCrtcReg(0x42) | 0x20;
		SetCrtcReg(0x42, d1);
	}
}

/*
3C4h index 12h W(R/W):  "Video PLL Data"                            (732/764)
bit  0-4  N1. Frequency divider. Stored as 1-31, actual value 3-33
	5-7  N2. Divides the frequency. 0: /1, 1: /2, 2: /4, 3: /8
	8-14  M.  Quotient. Stored as 1-127, actual value 3-129
Note: Frequency is (M/N1)/(1 << N2) *base frequency.M and N1 are the actual
	values, not the stored ones. Typically the base frequency is 14.318 MHz.
*/

static long GetS3VclkParam(long dotclock, int STACK_PTR *s_mul, int STACK_PTR *s_div, int STACK_PTR *s_p)
{
	int best_mul,best_div,best_p;
	long best_clock,best_vco;
	int pll_mul,pll_div,pll_p;

	CARD_LOG(("s3: clock range is %.2f - %.2f MHz\n", (double)s3_card->vco_min / 1000.0 / (1 << s3_card->p_max), (double)s3_card->vco_max / 1000.0 / (1 << s3_card->p_min) ));

	best_clock = -1;
	best_mul = 0;
	best_div = 0;
	best_p = 0;
	best_vco = 0;

	for(pll_div=s3_card->div_min;pll_div<=s3_card->div_max;++pll_div)
	for(pll_p=s3_card->p_min;pll_p<=s3_card->p_max;++pll_p)
	for(pll_mul=s3_card->mul_min;pll_mul<=s3_card->mul_max;++pll_mul) {
		long vco = s3_card->ref * 1000L * pll_mul / pll_div;
		long new_clock = vco / (1 << pll_p);

		if (vco < s3_card->vco_min*1000L && pll_p!=s3_card->p_max)
			continue;
		if (vco > s3_card->vco_max*1000L)
			continue;

		if (best_clock < 0 || labs(new_clock - dotclock) < labs(best_clock - dotclock)) {
			best_mul = pll_mul;
			best_div = pll_div;
			best_p = pll_p;
			best_clock = new_clock;
			best_vco = vco;
		}
	}

	if (best_clock < 0)
		return -1;

	if (best_vco < s3_card->vco_min*1000L)
		CARD_LOG(("s3: VCO clock %.2f is OUT of range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)s3_card->vco_min / 1000, (double)s3_card->vco_max / 1000));
	else
		CARD_LOG(("s3: VCO clock %.2f is in range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)s3_card->vco_min / 1000, (double)s3_card->vco_max / 1000));

	*s_mul = best_mul;
	*s_div = best_div;
	*s_p = best_p;

	CARD_LOG(("s3: ref:%.2f, mul:%d, div:%d, shift:%d\n",(double)s3_card->ref / 1000,best_mul,best_div,best_p));

	return best_clock;
}

static void SetS3VClock(int pll_mul, int pll_div, int pll_p)
{
	unsigned d0;

	if (s3_card->clock_type == S3_CLOCK_TRIO) {
		unsigned SR12,SR13;
		SR12 = ((pll_div-2) & 0x1F) | ((pll_p & 0x7) << 5);
		SR13 = ((pll_mul-2) & 0x7F);

		SetSeqReg(0x12, SR12);
		SetSeqReg(0x13, SR13);
	} else if (s3_card->clock_type == S3_CLOCK_VIRGEGX2) {
		unsigned SR12,SR13,SR29,ndiv;

		ndiv = ((pll_div-2) & 0x1F) | ((pll_p & 0x7) << 5);

		SR29 = ndiv >> 7;
		SR12 = (ndiv & 0x1f) | ((ndiv & 0x60) << 1);
		SR13 = ((pll_mul-2) & 0x7F);

		SetSeqReg(0x12, SR12);
		SetSeqReg(0x13, SR13);
		SetSeqReg(0x29, SR29);
	} else if (s3_card->clock_type == S3_CLOCK_AURORA) {
		unsigned SR12,SR13;
		SR12 = ((pll_div-2) & 0x3F) | ((pll_p & 0x7) << 6);
		SR13 = ((pll_mul-2) & 0x7F);

		SetSeqReg(0x12, SR12);
		SetSeqReg(0x13, SR13);
	}

	d0 = GetSeqReg(0x15) & ~0x21;
	d0 = d0 & ~(1 << 4); /* disable sequence divider */

	SetSeqReg(0x15, d0 | 0x03);
	SetSeqReg(0x15, d0 | 0x23);
	SetSeqReg(0x15, d0 | 0x03);

	SetS3PLL();
}
