/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "compil.h"
#include "vbe3.h"
#include "card.h"

/***************************************************************************/
/* VBE3 */

const char* driverVBE3(void) {
	return "Generic VBE3 board";
}

int detectVBE3(void) {
	return 1;
}

int setVBE3(void) {
	STACK_DECL CRTCPARAM cp;
	int hde; /* horizontal display end */
	int vde; /* vertical display end */
	unsigned x_mul; /* x_mul = Horizontal Display End * 8 / WIDTH */
	unsigned y_mul; /* y_mul = Verical Display End / HEIGHT */

	if (!GetMode15KHz(&cp))
		return 0;

	hde = GetCrtcReg(0x1);
	hde += 1;
	if (hde == cp.original_width) {
		x_mul = 8;
	} else if (2*hde == cp.original_width) {
		x_mul = 4;
	} else if (4*hde == cp.original_width) {
		x_mul = 2;
	} else if (8*hde == cp.original_width) {
		x_mul = 1;
	} else {
		return 0;
	}

	vde = GetCrtcReg(0x12);
	vde |= (((unsigned)GetCrtcReg(0x7)) << 7) & 0x100;
	vde |= (((unsigned)GetCrtcReg(0x7)) << 3) & 0x200;
	vde += 1;
	if (vde == cp.original_height) {
		y_mul = 1;
	} else if (vde == 2*cp.original_height) {
		y_mul = 2;
	} else if (vde == 4*cp.original_height) {
		y_mul = 4;
	} else {
		return 0;
	}

	if (cp.interlace)
		y_mul *= 2;
	if (cp.doublescan)
		y_mul /= 2;

	cp.HDisp *= x_mul;
	cp.HBStart *= x_mul;
	cp.HSStart *= x_mul;
	cp.HSEnd *= x_mul;
	cp.HBEnd *= x_mul;
	cp.HTotal *= x_mul;

	cp.VDisp *= y_mul;
	cp.VBStart *= y_mul;
	cp.VSStart *= y_mul;
	cp.VSEnd *= y_mul;
	cp.VBEnd *= y_mul;
	cp.VTotal *= y_mul;

	DisableGenerateSignal();
	SetCrtcVGARegister(&cp);
	SetHVPolarity(cp.hpolarity,cp.vpolarity);
	EnableGenerateSignal();

	return 1;
}

void resetVBE3(void) {
}

