/*
 * This file is part of the AdvanceVBE project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "lib.h"

/***************************************************************************/
/* Standard */

int abs(int v) {
	if (v<0)
		return -v;
	else
		return v;
}

long labs(long v) {
	if (v<0)
		return -v;
	else
		return v;
}

void _pascal memcpy(void far * dst, const void far * src, unsigned count) {
	asm push ds
	asm cld
	asm les di,[dst]
	asm lds si,[src]
	asm mov cx,[count]
	asm rep movsb
	asm pop ds
}

void _pascal memset(void far * dst, unsigned src, unsigned count) {
	asm push ds
	asm cld
	asm les di,[dst]
	asm mov al,byte ptr [src]
	asm mov cx,[count]
	asm rep stosb
	asm pop ds
}

int _pascal strcmp(const char far * s1, const char far * s2) {
	while (*s1 && *s1==*s2) {
	++s1;
	++s2;
	}
	if (((unsigned char)*s1)<((unsigned char)*s2)) return -1;
	if (((unsigned char)*s1)>((unsigned char)*s2)) return 1;
	return 0;
}

int _pascal memcmp(const void far * s1, const void far * s2,unsigned count) {
	unsigned i;
	for(i=0;i<count;++i) {
		if ((((unsigned char far *)s1)[i])<(((unsigned char far *)s2)[i])) return -1;
		if ((((unsigned char far *)s1)[i])>(((unsigned char far *)s2)[i])) return 1;
	}
	return 0;
}

void _pascal strncpy(char far *s1, const char far *s2, unsigned maxlen) {
	unsigned i = 0;
	while (i<maxlen && s2[i]) {
		s1[i] = s2[i];
		++i;
	}
	if (i<maxlen)
		s1[i] = 0;
}

unsigned _pascal strlen(const char far * s1) {
	unsigned i = 0;
	while (s1[i]) {
		++i;
	}
	return i;
}

void _pascal strcpy(char far* s1, const char far* s2) {
	unsigned i = 0;
	while (s2[i]) {
		s1[i] = s2[i];
		++i;
	}
	s1[i] = 0;
}

long _pascal strtol(const char far* s, unsigned radix) {
	long v = 0;
	int sign = 0;
	if (*s=='-') {
		++s;
		sign = 1;
	}
	while (*s>='0' && *s<='9' || *s>='a' && *s<='f' || *s>='A' && *s<='F') {
		v *= radix;
		if (*s>='0' && *s<='9')
			v += *s - '0';
		else if (*s>='a' && *s<='f')
			v += *s - 'a' + 10;
		else
			v += *s - 'A' + 10;
		++s;
	}
	if (sign)
		v = -v;
	return v;
}

long _pascal strtold(const char far* s, long multiplier) {
	long v = 0;
	int sign = 0;
	if (*s=='-') {
		++s;
		sign = 1;
	}
	while (*s>='0' && *s<='9') {
		v *= 10;
		v += *s - '0';
		++s;
	}
	v *= multiplier;
	if (*s=='.') {
		++s;
		while (*s>='0' && *s<='9') {
			multiplier /= 10;
			v += (*s - '0')*multiplier;
			++s;
		}
	}
	if (sign)
		v = -v;
	return v;
}

/*
double _pascal strtod(const char far* s) {
	double v = 0;
	int sign = 0;
	if (*s=='-') {
		++s;
		sign = 1;
	}
	while (*s>='0' && *s<='9') {
		v *= 10;
		v += *s - '0';
		++s;
	}
	if (*s=='.') {
		double b = 1;
		++s;
		while (*s>='0' && *s<='9') {
			b /= 10;
			v += (*s - '0')*b;
			++s;
		}
	}
	if (sign)
		v = -v;
	return v;
}
*/

/***************************************************************************/
/* Hardware */

/*
// Enter a critical section
// return:
//   ==0 no, enter
//   ==1 ok, enter
int _pascal sem_enter(uint16 far* psem) {
	asm les bx,[psem]
	asm mov ax,1
	asm xchg ax,[es:bx]
	asm cmp ax,0
	asm je ok
	return 0;
ok:
	return 1;
}

// leave from a critical section
void _pascal sem_leave(uint16 far* psem) {
	*psem = 0;
}

// Disable interrupt and enter a critical section
// return:
//   ==0 no, NOT enter
//   !=0 ok, enter
int _pascal region_enter(struct region far* pregion) {
	uint16 flags = disable();
	if (sem_enter(&pregion->sem)) {
		pregion->flags = flags;
		return 1;
	} else {
		restore(flags);
		return 0;
	}
}

void _pascal region_leave(struct region far* pregion) {
	uint16 flags = pregion->flags;
	sem_leave(&pregion->sem);
	restore(flags);
}

// Disable interrupt
// return:
//   flags to be passed to enable
uint16 _pascal disable() {
	asm pushf
	asm pop ax
	asm cli
	return _AX;
}

// Abilita interrupt
// return:
//   flags to be passed to enable
uint16 _pascal enable() {
	asm pushf
	asm pop ax
	asm sti
	return _AX;
}

// Reimposta i flags
void _pascal restore(uint16 flags) {
	asm mov ax,[flags]
	asm push ax
	asm popf
}
*/

void _pascal outportb(int port, uint8 value) {
	asm mov al,[value]
	asm mov dx,[port]
	asm out dx,al
}

uint8 _pascal inportb(int port) {
	uint8 value;
	asm mov dx,[port]
	asm in al,dx
	asm mov [value],al
	return value;
}

void _pascal outportw(int port, uint16 value) {
	asm mov ax,[value]
	asm mov dx,[port]
	asm out dx,ax
}

uint16 _pascal inportw(int port) {
	uint16 value;
	asm mov dx,[port]
	asm in ax,dx
	asm mov [value],ax
	return value;
}

void _pascal outportl(int port, uint32 value) {
	asm mov eax,[value]
	asm mov dx,[port]
	asm out dx,eax
}

uint32 _pascal inportl(int port) {
	uint32 value;
	asm mov dx,[port]
	asm in eax,dx
	asm mov [value],eax
	return value;
}

/***************************************************************************/
/* Timer */

/*
static unsigned near readtimer() {
	asm pushf
	asm cli
	asm mov  al,0h
	asm out  43h,al
	asm nop
	asm in   al,40h
	asm mov  bl,al
	asm nop
	asm in   al,40h
	asm mov  bh,al
	asm not  bx
	asm popf
	return _BX;
}

#define DELAY_MULTIPLIER 1193UL

void _pascal delay( unsigned milliseconds ) {
	unsigned long stop;
	unsigned cur, prev;

	stop = (prev = readtimer()) + (milliseconds * DELAY_MULTIPLIER);

	while ((cur = readtimer()) < stop) {
		if (cur < prev) {
			if (stop < 0x10000L)
				break;
			stop -= 0x10000L;
		}
		prev = cur;
	}
}
*/

/***************************************************************************/
/* Sound */

/*
// Outputs a sound of a specified frequency to the speaker
void _pascal sound(unsigned frequency) {
	asm     mov     bx,  frequency
	asm     mov     ax,  34DDh
	asm     mov     dx,  0012h
	asm     cmp     dx,  bx
	asm     jnb     stop
	asm     div     bx
	asm     mov     bx,  ax
	asm     in      al,  61h
	asm     test    al,  3
	asm     jne     j1
	asm     or      al,  3
	asm     out     61h, al
	asm     mov     al,  0B6h
	asm     out     43h, al
	j1:
	asm     mov     al,  bl
	asm     out     42h, al
	asm     mov     al,  bh
	asm     out     42h, al
	stop: ;
}

// Turns the speaker off
void _pascal nosound() {
	asm     in      al,61H
	asm     and     al, 0fcH
	asm     out     61H, al
}
*/

/***************************************************************************/
/* BIOS */

void _pascal cputc(unsigned char c) {
	asm mov ah,0Eh
	asm mov al,[c]
	asm mov bx,7
	asm int 10h
}

void _pascal cputs(const char far * s) {
	while (*s) cputc(*s++);
}

void _pascal cputu(unsigned n, unsigned w, char pad, unsigned base) {
	char s[32];
	unsigned l = 0;
	while (n) {
		unsigned d = n % base;
		s[l++] = d < 10 ? '0'+d : 'a'+d-10;
		n /= base;
	}
	if (!l) s[l++] = '0';
	while (l<w) s[l++] = pad;
	while (l) cputc(s[--l]);
}

/*
void _pascal cputi(int n, unsigned w, char pad, unsigned base) {
	char s[32];
	unsigned l = 0;
	int sign = n<0;
	if (sign) n = -n;
	while (n) {
		unsigned d = n % base;
		s[l++] = d < 10 ? '0'+d : 'a'+d-10;
		n /= base;
	}
	if (!l) s[l++] = '0';
	if (sign) s[l++] = '-';
	while (l<w) s[l++] = pad;
	while (l) cputc(s[--l]);
}
*/

/*
--------V-1003-------------------------------
INT 10 - VIDEO - GET CURSOR POSITION AND SIZE
	AH = 03h
	BH = page number
			0-3 in modes 2&3
			0-7 in modes 0&1
			0 in graphics modes
Return: AX = 0000h (Phoenix BIOS)
	CH = start scan line
	CL = end scan line
	DH = row (00h is top)
	DL = column (00h is left)
Notes:  a separate cursor is maintained for each of up to 8 display pages
	many ROM BIOSes incorrectly return the default size for a color display
		(start 06h, end 07h) when a monochrome display is attached
SeeAlso: AH=01h,AH=02h,AH=12h/BL=34h,MEM 0040h:0050h,MEM 0040h:0060h
*/
/*
unsigned _pascal wherex() {
	asm mov ah,3h
	asm mov bh,0
	asm int 10h
	asm mov dh,0
	asm mov ax,dx
	return _AX;
}

unsigned _pascal wherey() {
	asm mov ah,3h
	asm mov bh,0
	asm int 10h
	asm shr dx,8
	asm mov ax,dx
	return _AX;
}
*/

/*
--------V-1002-------------------------------
INT 10 - VIDEO - SET CURSOR POSITION
	AH = 02h
	BH = page number
			0-3 in modes 2&3
			0-7 in modes 0&1
			0 in graphics modes
	DH = row (00h is top)
	DL = column (00h is left)
SeeAlso: AH=03h,AH=05h,INT 60/DI=030Bh,MEM 0040h:0050h
*/
/*
void _pascal gotoxy(unsigned x, unsigned y) {
	asm mov ah,2h
	asm mov bh,0
	asm mov dl,[x]
	asm mov dh,[y]
	asm int 10h
}
*/

/***************************************************************************/
/* DOS */

uint16 errno;

uint16 _pascal pspseg() {
	asm mov ah,62h
	asm int 21h
	asm mov ax,bx
	return _AX;
}

/*
Format of environment block:
Offset  Size  Description (Table 0604)
	00h  N BYTEs first environment variable, ASCIZ string of form "var=value"
		N BYTEs second environment variable, ASCIZ string
	...
		N BYTEs last environment variable, ASCIZ string of form "var=value"
	BYTE  00h
---DOS 3.0+ ---
	uint16  number of strings following environment (normally 1)
		N BYTEs ASCIZ full pathname of program owning this environment
		other strings may follow
*/

// Return the environment segment
// return:
//   On old DOS can be 0
uint16 _pascal envseg() {
	return *((uint16 far *)MK_FP(pspseg(),0x2C));
}

const char far* _pascal env() {
	uint16 seg = envseg();
	if (seg)
		return (const char far *)MK_FP(seg, 0);
	else
		return 0;
}

// Release the environment segment
int _pascal freeenv() {
	return freemem( envseg() );
}

// Compute the environment size
// return:
//   0 if invalid
unsigned _pascal envsize() {
	unsigned i;
	const char far * e = env();
	if (!e) return 0;
	i = 0;
	while (i<=32767 && e[i]!=0) {
		while (i<=32767 && e[i]!=0) {
			++i;
		}
		++i;
	}
	++i;
	if (i>32767) return 0; // environment is limited to 32K
	return i;
}

/*
// Full pathname of program owning this environment
const char far* _pascal exec_pathname() {
	unsigned size = envsize();
	if (!size) return 0;
	const char far * e = env();
	return e+size+2;
}
*/

int _pascal freemem(uint16 memseg) {
	asm mov ax,[memseg]
	asm mov es,ax
	asm mov ah,49h
	asm int 21h
	asm jc error
	return 0;
error:
	asm mov [errno],ax
	return -1;
}

/*
int _pascal allocmem(unsigned size, uint16 far* segp) {
	asm mov ah,48h
	asm mov bx,size
	asm int 21h
	asm jc error
	asm les bx,[segp]
	asm mov [es:bx],ax
	return 0;
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
// dos put char
void _pascal putc(char c) {
	asm mov ah,0x02
	asm mov dl,c
	asm int 0x21
}

// dos put string
void _pascal puts(const char * s) {
	while (*s) putc(*(s++));
}
*/

/***************************************************************************/
/* File */

/*
int _pascal creat(const char far* file, unsigned attrib) {
	asm mov ah,03Ch
	asm mov cx,[attrib]
	asm push ds
	asm lds dx,[file]
	asm int 21h
	asm pop ds
	asm jc error
	return _AX;
error:
	asm mov [errno],ax
	return -1;
}
*/

int _pascal open(const char far* file, unsigned access) {
	asm mov ah,3Dh
	asm mov al,[byte ptr access]
	asm mov cl,0
	asm push ds
	asm lds dx,[file]
	asm int 21h
	asm pop ds
	asm jc error
	return _AX;
error:
	asm mov [errno],ax
	return -1;
}

int _pascal close(int handle) {
	asm mov ah,03Eh
	asm mov bx,[handle]
	asm int 21h
	asm jc error
	return 0;
error:
	asm mov [errno],ax
	return -1;
}

int _pascal read(int handle, void far * buf, int len) {
	asm mov ah,3Fh
	asm mov bx,[handle]
	asm mov cx,[len]
	asm push ds
	asm lds dx,[buf]
	asm int 21h
	asm pop ds
	asm jc error
	return _AX;
error:
	asm mov [errno],ax
	return -1;
}

int _pascal fgetc(int handle) {
	char c;
	if (read(handle,&c,1)<=0)
		return EOF;
	else
		return c;
}

char far* _pascal gets(int handle, char far * buf, int len) {
	char far * p = buf;
	while (len>1) {
		int c = fgetc(handle);
		if (c==EOF) {
			if (p==buf)
				return 0;
			else
				break;
		}
		if (c=='\n')
			break;
		if (c!='\r') {
			*p = c;
			++p;
			--len;
		}
	}
	*p = 0;
	return buf;
}

/*
int _pascal write(int handle,const void far * buf, int len) {
	asm mov ah,40h
	asm mov bx,[handle]
	asm mov cx,[len]
	asm push ds
	asm lds dx,[buf]
	asm int 21h
	asm pop ds
	asm jc error
	return _AX;
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
long _pascal lseek(int handle, long off, int fromwhere) {
	asm mov ah,42h
	asm mov al,[byte ptr fromwhere]
	asm mov bx,[word ptr handle]
	asm mov dx,[word ptr off]
	asm mov cx,[word ptr off+2]
	asm int 21h
	asm jc error
	return (long)((void _seg *)(_DX) + (void near *)(_AX));
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
long _pascal filelength(int handle) {
	long pos = lseek(handle,0,SEEK_CUR);
	if (pos>=0) {
		long end = lseek(handle,0,SEEK_END);
		if (end>=0)
			if (lseek(handle,pos,SEEK_SET)>=0)
				return end;
	}
	return -1;
}
*/

/*
int _pascal getftime(int handle, ftime far* ftimep) {
	asm mov ax,5700h
	asm mov bx,[handle]
	asm int 21h
	asm jc error
	asm les bx,[ftimep]
	asm mov [es:bx],cx
	asm mov [es:bx+2],dx
	return 0;
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
int _pascal setftime(int handle, const ftime far* ftimep) {
	asm mov ax,5701h
	asm les bx,[ftimep]
	asm mov cx,[es:bx]
	asm mov dx,[es:bx+2]
	asm mov bx,[handle]
	asm int 21h
	asm jc error
	return 0;
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
int _pascal getfattr(const char far *path, unsigned far *attribp) {
	asm mov ax,4300h
	asm push ds
	asm lds dx,[path]
	asm int 21h
	asm pop ds
	asm jc error
	asm les bx,[attribp]
	asm mov [es:bx],cx
	return 0;
error:
	asm mov [errno],ax
	return -1;
}
*/

/*
int _pascal setfattr(const char far *path, unsigned attrib) {
	asm mov ax,4301h
	asm mov cx,[attrib]
	asm push ds
	asm lds dx,[path]
	asm int 21h
	asm pop ds
	asm jc error
	return 0;
error:
	asm mov [errno],ax
	return -1;
}
*/

/***************************************************************************/
/* Date/Time */

/*
// Converte un numero giuliano in data
void _pascal jul2date(long jul, date far* datep) {
	long day, month, year, tmp;

	tmp = jul-1721119L;
	year = (4*tmp-1)/146097L;
	tmp = 4*tmp-1-146097L*year;
	day = tmp/4;
	tmp = (4*day+3)/1461;
	day = 4*day+3-1461*tmp;
	day = (day+4)/4;
	month = (5*day-3)/153;
	day = 5*day-3-153*month;
	day = (day+5)/5;
	year = 100*year+tmp;
	datep->da_day = (int)day;
	datep->da_year = (int)year;

	if ((int)month < 10)
		datep->da_mon = (int)month+3;
	else {
		datep->da_mon = (int)month-9;
		datep->da_year++;
	}
}
*/

/*
// Converte una data (giorno, mese, anno) in numero giuliano
// in:
//   day giorno della data da convertire (1-31)
//   month mese della data da convertire (1-12).
//   year anno della data da convertire (19xx|2xxx)
long _pascal date2jul(const date far* datep) {
	long cent, day, month, year;

	if (datep->da_mon > 2) {
		month = (long)(datep->da_mon-3);
		year = (long)datep->da_year;
	} else {
		month = (long)(datep->da_mon+9);
		year = (long)(datep->da_year-1);
	}

	cent = (year/100);
	day = year-(100*cent);

	return ((146097L*cent)/4L+(1461L*day)/4L+(153L*month+2)/5L+1721119L+(long)datep->da_day);
}
*/

/*
void _pascal gettime(time far *timep) {
	asm mov ah,2Ch
	asm int 21h
	asm les bx,[timep]
	asm mov [byte ptr es:bx],cl
	asm mov [byte ptr es:bx+1],ch
	asm mov [byte ptr es:bx+2],dl
	asm mov [byte ptr es:bx+3],dh
}
*/

/*
void _pascal getdate(date far *datep) {
	asm mov ah,2Ah
	asm int 21h
	asm les bx,[datep]
	asm mov [word ptr es:bx],cx
	asm mov [byte ptr es:bx+2],dl
	asm mov [byte ptr es:bx+3],dh
}
*/

/***************************************************************************/
/* Int */

/*
// Call an IRQ
void _pascal irq_call(void far * intr) {
	asm pushf
	asm call [dword ptr intr]
}
*/

void far* _pascal getvect(int intr) {
	asm mov ah,035h
	asm mov al,[byte ptr intr]
	asm int 21h
	asm mov ax,bx
	asm mov dx,es
	return MK_FP(_DX,_AX);
}

void _pascal setvect(int intr, void far * func) {
	asm mov ah,025h
	asm mov al,[byte ptr intr]
	asm push ds
	asm lds dx, dword ptr func
	asm int 21h
	asm pop ds
}

void _pascal regs_16_preset(regs_16 far* regs) {
	regs->x.es = _DS;
}

void _pascal regs_32_preset(regs_32 far* regs) {
	regs->d.es = _DS;
}

void _pascal regs_16_zero(regs_16 far* regs) {
	memset(regs,0,sizeof(regs_16));
	regs_16_preset(regs);
}

void _pascal regs_32_zero(regs_32 far* regs) {
	memset(regs,0,sizeof(regs_32));
	regs_32_preset(regs);
}
